import datetime
from typing import Dict, Sequence

from transmart_loader.transmart import Concept, TreeNode, ValueType, ConceptNode, TreeNodeMetadata

from csr.csr import SubjectEntity, StudyEntity
from csr2transmart.ontology_config import TreeNode as OntologyConfigTreeNode, OntologyConfigValidationException


class OntologyMapper:
    """
    Map ontology config to tree nodes and concepts
    """
    def __init__(self, top_tree_node: str):
        self.top_tree_node = top_tree_node
        self.concept_code_to_concept: Dict[str, Concept] = {}

    @staticmethod
    def type_to_value_type(field_type) -> ValueType:
        if field_type in ['integer', 'number']:
            return ValueType.Numeric
        if field_type in ['date']:
            return ValueType.Date
        else:
            return ValueType.Categorical

    @staticmethod
    def entity_name_to_subject_dimension_value(entity_name: str) -> str:
        if entity_name == 'Individual':
            return 'patient'
        else:
            return entity_name

    @staticmethod
    def is_concept_node(node: OntologyConfigTreeNode) -> bool:
        return node.concept_code is not None

    def get_concept_type(self, entity_name: str, entity_field_name: str) -> ValueType:
        entity_types = list(SubjectEntity.__args__)
        entity_types.extend(list(StudyEntity.__args__))
        entity_type: SubjectEntity = list(filter(lambda x: x.schema()['title'] == entity_name, entity_types))[0]
        field_type = entity_type.schema()['properties'][entity_field_name]['type']
        return self.type_to_value_type(field_type)

    def map_concept_node(self, node: OntologyConfigTreeNode) -> ConceptNode:
        entity_name, entity_field_name = node.concept_code.split('.')
        concept_path = '\\\\CSR\\\\' + node.concept_code
        concept_type = self.get_concept_type(entity_name, entity_field_name)
        concept = Concept(node.concept_code, node.name, concept_path, concept_type)
        self.concept_code_to_concept[node.concept_code] = concept

        metadata_value: Dict[str, str] = {
            'subject_dimension': self.entity_name_to_subject_dimension_value(entity_name)}
        concept_node = ConceptNode(concept)
        concept_node.metadata = TreeNodeMetadata(metadata_value)
        return concept_node

    def map_nodes(self, nodes: Sequence[OntologyConfigTreeNode], parent_node: TreeNode):
        for node in nodes:
            if isinstance(node, dict):
                node = OntologyConfigTreeNode(**node)
            elif not isinstance(node, OntologyConfigTreeNode):
                raise OntologyConfigValidationException(f'Invalid ontology node: {node}')

            if self.is_concept_node(node):
                parent_node.add_child(self.map_concept_node(node))
            else:
                intermediate_node = TreeNode(node.name)
                self.map_nodes(node.children, intermediate_node)
                parent_node.add_child(intermediate_node)

    def map(self, src_nodes: Sequence[OntologyConfigTreeNode]) -> Sequence[TreeNode]:
        date = datetime.datetime.now().strftime('%d-%m-%Y')
        metadata = TreeNodeMetadata({'Load date': date})
        top_node = TreeNode(self.top_tree_node, metadata)
        self.map_nodes(src_nodes, top_node)
        return [top_node]



import logging
import os
from typing import Set

from csr.csr import CentralSubjectRegistry
from sources2csr.ngs import NGS
from sources2csr.ngs_maf_reader import NgsMafReader
from sources2csr.ngs_reader import NgsReader
from sources2csr.ngs_seg_reader import NgsSegReader
from sources2csr.ngs_txt_reader import NgsTxtReader


logger = logging.getLogger(__name__)


def read_ngs_files(ngs_dir: str) -> Set[NGS]:
    """ Reads NGS files inside input_dir in 3 different formats: `.txt`, `.maf.gz` and `.seg`.

    :param ngs_dir: NGS files directory
    :return: Set of NGS objects
    """
    ngs_data = set()
    txt_reader = NgsTxtReader(ngs_dir)
    maf_reader = NgsMafReader(ngs_dir)
    seg_reader = NgsSegReader(ngs_dir)

    for filename in NgsReader.list_files(ngs_dir):
        # logger.debug('Parsing {} data file from {}'.format(filename, directory))
        if filename.endswith('maf.gz'):
            ngs_data |= set(maf_reader.read_data(filename))
        if filename.endswith('seg'):
            ngs_data |= set(seg_reader.read_data(filename))
        elif filename.endswith('_all_data_by_genes.txt') or filename.endswith('_all_thresholded.by_genes.txt'):
            ngs_data |= set(txt_reader.read_data(filename))
    return ngs_data


def read_ngs_data(input_dir: str) -> Set[NGS]:
    ngs_dir = os.path.join(input_dir, 'NGS')
    if not os.path.isdir(ngs_dir):
        logger.info('No NGS data found.')
        return set()
    logger.info('Reading NGS data.')
    return read_ngs_files(ngs_dir)


def add_ngs_data(subject_registry: CentralSubjectRegistry, input_dir: str) -> CentralSubjectRegistry:
    """Add library_strategy aggregates and analysis_type to biomaterials based on the NGS data
    :param subject_registry: Central Subject Registry
    :param input_dir: input directory that contains the directory with NGS input files
    :return: updated Central Subject Registry
    """
    ngs_data = read_ngs_data(input_dir)
    if subject_registry.biomaterials and ngs_data:
        for biomaterial in subject_registry.biomaterials:
            biomaterial.analysis_type = []
            biomaterial.library_strategy = []
            for ngs in ngs_data:
                if biomaterial.biomaterial_id == ngs.biomaterial_id and \
                   biomaterial.src_biosource_id == ngs.biosource_id:
                    if ngs.analysis_type is not None:
                        biomaterial.analysis_type.append(ngs.analysis_type.value)
                    biomaterial.library_strategy.append(ngs.library_strategy.value)
    return subject_registry



#!/usr/bin/env python3

# Code to transform clinical data
# Author: Sander Tan, The Hyve

import logging
import os
import re
import sys

import numpy as np
import pandas as pd

from csr.csr import CentralSubjectRegistry
from .create_metafile import create_meta_content

logger = logging.getLogger(__name__)
logger.name = logger.name.rsplit('.', 1)[1]

# Rename column headers in patient clinical data
RENAME_CLINICAL_DATA_HEADER = {'INDIVIDUAL_ID': 'PATIENT_ID'}
# Force some datatypes to be remappad to STRING
FORCE_STRING_LIST = ['CENTER_TREATMENT', 'CID', 'DIAGNOSIS_ID', 'GENDER', 'IC_DATA', 'IC_LINKING_EXT', 'IC_MATERIAL',
                     'IC_TYPE', 'INDIVIDUAL_STUDY_ID', 'TOPOGRAPHY', 'TUMOR_TYPE']


def create_clinical_header(df):
    """ Function to create header in cBioPortal format for clinical data
    """

    # Create attribute name and attribute description
    header_data = pd.DataFrame(data=df.columns)
    header_data[1] = header_data

    # Data type header line
    type_values = []
    for i in df.dtypes:
        if np.issubdtype(i, np.number):
            type_values.append('NUMBER')
        else:
            type_values.append('STRING')

    # Create attribute type and attribute priority
    header_data[2] = type_values
    header_data[3] = 1

    # Transpose to put in correct format
    header_data = header_data.transpose()

    # Transform to dataframe
    header_data = pd.DataFrame(header_data)
    header_data.columns = df.columns

    # TODO: Major - Check FORCE_STRING_LIST assumption --> make configurable?
    # Set datatype of specific columns before after header
    for string_column_name in FORCE_STRING_LIST:
        if string_column_name in header_data.columns:
            header_data.loc[2, string_column_name] = 'STRING'

    # Adding hash # to the first column
    header_data.iloc[:, 0] = '#' + header_data.iloc[:, 0].astype(str)

    return header_data


def to_sample_data(biosource_data, biomaterial_data, diagnosis_data) -> pd.DataFrame:
    """ Merge diagnoses, biosources and biomaterials into samples data frame
    """
    diagnosis_biosource_data = pd.merge(biosource_data, diagnosis_data, how='left', on=['DIAGNOSIS_ID'])
    sample_data = pd.merge(biomaterial_data, diagnosis_biosource_data, how='left',
                           left_on=['SRC_BIOSOURCE_ID'], right_on=['BIOSOURCE_ID'])
    # remove suffixes for biomaterials and biosources
    if 'SRC_BIOSOURCE_ID_x' in sample_data.columns:
        x_ = sample_data['SRC_BIOSOURCE_ID_x']
        y_ = sample_data['SRC_BIOSOURCE_ID_y']
        sample_data['SRC_BIOSOURCE_ID'] = x_.combine_first(y_)
        sample_data = sample_data.drop(columns=['SRC_BIOSOURCE_ID_x', 'SRC_BIOSOURCE_ID_y'])
    # remove suffixes for patients
    if 'PATIENT_ID_x' in sample_data.columns:
        x_ = sample_data['PATIENT_ID_x']
        y_ = sample_data['PATIENT_ID_y']
        sample_data['PATIENT_ID'] = x_.combine_first(y_)
        sample_data = sample_data.drop(columns=['PATIENT_ID_x', 'PATIENT_ID_y'])
    sample_data['SAMPLE_ID'] = sample_data['BIOSOURCE_ID'] + "_" + sample_data['BIOMATERIAL_ID']
    return sample_data


def subject_registry_to_sample_data_df(subject_registry: CentralSubjectRegistry) -> pd.DataFrame:
    diagnosis_data = pd.DataFrame.from_records([s.__dict__ for s in subject_registry.diagnoses])
    biosource_data = pd.DataFrame.from_records([s.__dict__ for s in subject_registry.biosources])
    biomaterial_data = pd.DataFrame.from_records([s.__dict__ for s in subject_registry.biomaterials])

    diagnosis_data.columns = [x.upper() for x in diagnosis_data.columns]
    biosource_data.columns = [x.upper() for x in biosource_data.columns]
    biomaterial_data.columns = [x.upper() for x in biomaterial_data.columns]

    diagnosis_data.rename(columns=RENAME_CLINICAL_DATA_HEADER, inplace=True)
    biosource_data.rename(columns=RENAME_CLINICAL_DATA_HEADER, inplace=True)

    return to_sample_data(biosource_data, biomaterial_data, diagnosis_data)


def subject_registry_to_patient_data_df(subject_registry: CentralSubjectRegistry) -> pd.DataFrame:
    patient_data = pd.DataFrame.from_records([s.__dict__ for s in subject_registry.individuals])

    patient_data.columns = [x.upper() for x in patient_data.columns]
    patient_data.rename(columns=RENAME_CLINICAL_DATA_HEADER, inplace=True)

    return patient_data


def check_integer_na_columns(numerical_column):
    """ Integer columns containing NA values are converted to Float in pandas
        This function checks for a column if it only contains integers (.0)
        In this case we attempt to convert all columns, not just the columns
        that contain an NA value, because we sliced the dataframes from a bigger
        dataframe that could contain NA values.
    """
    for i in numerical_column:
        if str(i) != 'nan':
            decimal_value = str(i).split('.')[1]
            # If float is other than 0, we know for sure it's not an integer column
            if float(decimal_value) != 0:
                return False
    # If loop ends by only encountering .0, return true
    return True


def fix_integer_na_columns(df):
    """ Integer columns containing NA values are converted to Float in pandas
        This function selects these columns and converts them to back to integers
    """

    # Select data with floats
    numerical_data = df.select_dtypes(include=[np.float64])

    # Round values to 4 decimals
    # clinical_data.loc[:, numerical_data.columns] = np.round(numerical_data, 4)

    # Select columns which contain integers with NA values (numpy has converted them to floats)
    integer_na_columns = numerical_data.columns[numerical_data.apply(check_integer_na_columns, axis=0)]

    # Convert integer columns with NA to string(integer) columns
    if len(integer_na_columns) > 0:
        df.loc[:, integer_na_columns] = df.loc[:, integer_na_columns].applymap(
            lambda x: str(x).split('.')[0].replace('nan', 'NA'))
    return df


def modify_clinical_data_column_names(clinical_data_df):
    # Remove symbols from attribute names and make them uppercase
    clinical_data_df = clinical_data_df.rename(columns=lambda s: re.sub('[^0-9a-zA-Z_]+', '_', s))
    clinical_data_df = clinical_data_df.rename(columns=lambda x: x.strip('_'))
    return clinical_data_df


def modify_clinical_data_column_values(clinical_data_df):
    # Fix integers that are converted to floats due to NA in column
    if len(clinical_data_df.select_dtypes(include=[np.float64]).columns) > 0:
        clinical_data_df = fix_integer_na_columns(clinical_data_df)
    # Check if column names are unique.
    # In case of duplicates, rename them manually before or after header creation
    # TODO: Nice to have - Extract mapping dictionaries to external config, now they are hardcoded
    if not len(set(clinical_data_df.columns.tolist())) == len(clinical_data_df.columns.tolist()):
        logger.error('Attribute names not unique, not writing data_clinical. Rename them using the remap dictionary.')
        sys.exit(1)
    # Drop duplicate rows, this does not check for duplicate sample ID's
    cd_row_count = clinical_data_df.shape[0]
    clinical_data_df = clinical_data_df.loc[clinical_data_df.astype(str).drop_duplicates(keep='first').index]
    logger.debug('Found and dropped {} duplicates in the clinical data'
                 .format(cd_row_count - clinical_data_df.shape[0]))
    return clinical_data_df


def transform_patient_clinical_data(subject_registry: CentralSubjectRegistry) -> [pd.DataFrame, pd.DataFrame]:
    patient_data_df = subject_registry_to_patient_data_df(subject_registry)

    # Remove empty columns
    patient_data_df.dropna(axis=1, how='all', inplace=True)
    # Create header
    patient_data_header = create_clinical_header(patient_data_df)
    patient_data_df = modify_clinical_data_column_names(patient_data_df)
    patient_data_df = modify_clinical_data_column_values(patient_data_df)

    return patient_data_df, patient_data_header


def transform_sample_clinical_data(subject_registry: CentralSubjectRegistry) -> [pd.DataFrame, pd.DataFrame]:
    sample_data_df = subject_registry_to_sample_data_df(subject_registry)

    # Remove empty columns
    sample_data_df.dropna(axis=1, how='all', inplace=True)
    # Create header
    sample_data_header = create_clinical_header(sample_data_df)
    sample_data_df = modify_clinical_data_column_names(sample_data_df)
    sample_data_df = modify_clinical_data_column_values(sample_data_df)

    return sample_data_df, sample_data_header


def write_clinical(clinical_data, clinical_header, clinical_type, output_dir, study_id):
    # Writing clinical patient file
    clinical_filename = os.path.join(output_dir, 'data_clinical_{}.txt'.format(clinical_type))
    clinical_header.to_csv(clinical_filename, sep='\t', index=False, header=False, mode='w')
    clinical_data.to_csv(clinical_filename, sep='\t', index=False, header=True, mode='a')
    # Set clinical type for meta file
    if clinical_type == 'sample':
        meta_datatype = 'SAMPLE_ATTRIBUTES'
    elif clinical_type == 'patient':
        meta_datatype = 'PATIENT_ATTRIBUTES'
    else:
        logger.error("Unknown clinical data type")
        sys.exit(1)
    # Create meta file
    meta_filename = os.path.join(output_dir, 'meta_clinical_{}.txt'.format(clinical_type))
    create_meta_content(file_name=meta_filename,
                        cancer_study_identifier=study_id,
                        genetic_alteration_type='CLINICAL',
                        datatype=meta_datatype,
                        data_filename='data_clinical_{}.txt'.format(clinical_type))



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class LibraryStrategy(Enum):
    """
    Type of library strategy (sequencing type)
    """
    SNV = "SNV"
    CNV = "CNV"


class AnalysisType(Enum):
    """
    Type of analysis
    """
    WGS = "WGS"
    WXS = "WXS"


class NGS(BaseModel):
    biosource_id: str
    biomaterial_id: str
    library_strategy: LibraryStrategy
    analysis_type: Optional[AnalysisType]

    def __hash__(self):
        return hash(('biosource_id', self.biosource_id,
                     'biomaterial_id', self.biomaterial_id,
                     'library_strategy', self.library_strategy,
                     'analysis_type', self.analysis_type))



#!/usr/bin/env python3

# Code to transform PMC processed data to cBioPortal staging files
# Author: Sander Tan, The Hyve

import csv
import errno
import gzip
import logging
import os
import shutil
import sys
import time
from typing import List, Set, Optional

import click
import pandas as pd
from csr.logging import setup_logging

from csr.csr import CentralSubjectRegistry
from csr.exceptions import DataException
from csr.subject_registry_reader import SubjectRegistryReader
from .create_caselist import create_caselist
from .create_metafile import create_meta_content
from .transform_clinical import write_clinical, transform_patient_clinical_data, transform_sample_clinical_data

logger = logging.getLogger(__name__)

# Define study properties
STUDY_ID = 'csr'
NAME = "Central Subject Registry"
NAME_SHORT = "Central Subject Registry"
DESCRIPTION = 'Transformed to cBioPortal format on: %s' % time.strftime("%d-%m-%Y %H:%M")
TYPE_OF_CANCER = 'mixed'


def prepare_output_directory(output_dir: str):
    # Remove old output directory and recreate to ensure all NGS data is
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    try:
        os.makedirs(output_dir, exist_ok=False)
    except OSError as oe:
        # Catch the error if the folder already exists as this indicates shutil.rmtree did something wrong
        if oe.errno == errno.EEXIST:
            logger.error('Failed to create the output directory {} as it already exists.\
             It was not cleared successfully, Error {}'.format(output_dir, oe))
        raise oe


def process_clinical_data(input_dir: str, output_dir: str) -> List[str]:
    """
    Reads subject registry data from input_dir and transforms the data
    to clinical data files for cBioPortal.
    Writes the generated data files to output_dir.
    Returns the list of sample identifiers in the clinical data.
    """
    # Clinical data
    subject_registry_reader = SubjectRegistryReader(input_dir)
    subject_registry: CentralSubjectRegistry = subject_registry_reader.read_subject_registry()

    # Transform patient file
    patient_clinical_data, patient_clinical_header = transform_patient_clinical_data(subject_registry)
    write_clinical(patient_clinical_data, patient_clinical_header, 'patient', output_dir, STUDY_ID)

    # Transform sample file
    sample_clinical_data, sample_clinical_header = transform_sample_clinical_data(subject_registry)
    write_clinical(sample_clinical_data, sample_clinical_header, 'sample', output_dir, STUDY_ID)

    return sample_clinical_data['SAMPLE_ID'].unique().tolist()


def process_mutation_data(ngs_dir: str, output_dir: str, clinical_sample_ids: List[str]) -> Set[str]:
    """
    Reads data from all MAF files in ngs_dir and creates a combined mutation data file
    and the meta and caselist files for the mutation data.

    Returns the list of sample identifiers found in the mutation data.
    An exception is raised when any of the sample identifiers is not in the clinical_sample_ids list.
    """
    mutation_samples = combine_maf(ngs_dir, os.path.join(output_dir, 'data_mutations.maf'))

    if mutation_samples:
        # Create meta file
        meta_filename = os.path.join(output_dir, 'meta_mutations.txt')
        create_meta_content(meta_filename, cancer_study_identifier=STUDY_ID,
                            genetic_alteration_type='MUTATION_EXTENDED', datatype='MAF',
                            stable_id='mutations', show_profile_in_analysis_tab='true',
                            profile_name='Mutations', profile_description='Mutation data',
                            data_filename='data_mutations.maf', variant_classification_filter='',
                            swissprot_identifier='accession')

        # Create case list
        create_caselist(output_dir=output_dir, file_name='cases_sequenced.txt',
                        cancer_study_identifier=STUDY_ID,
                        stable_id='%s_sequenced' % STUDY_ID, case_list_name='Sequenced samples',
                        case_list_description='All sequenced samples',
                        case_list_category='all_cases_with_mutation_data',
                        case_list_ids="\t".join(mutation_samples))

        # Test for samples in MAF files that are not in clinical data
        if not set(clinical_sample_ids).issuperset(mutation_samples):
            logger.error("Found samples in MAF files that are not in clinical data: {}".format(
                ", ".join(mutation_samples.difference(set(clinical_sample_ids)))))
            raise DataException('Found samples in MAF files that are not in clinical data')
    return mutation_samples


def process_cna_files(ngs_dir: str, output_dir: str, clinical_sample_ids: List[str]) -> List[str]:
    """
    Reads CNA data files (segmented, continuous and discrete) from ngs_dir, copies the files,
    drops and renames certain columns, and writes meta and case list files.

    Returns list of CNA sample identifiers.
    An exception is raised when any of the sample identifiers is not in the clinical_sample_ids list.
    """
    # Select all non-hidden files
    study_files = []
    for study_file in os.listdir(ngs_dir):
        if not study_file.startswith('.'):
            study_files.append(study_file)

    # Create sample list, required for cnaseq case list
    cna_samples = []

    # Transform CNA data files
    for study_file in study_files:
        if study_file.endswith('sha1'):
            continue

        # CNA Segment data
        if study_file.split('.')[-1] == 'seg':
            logger.debug('Transforming segment data: %s' % study_file)

            output_file = 'data_cna_segments.seg'

            # Read file and replace header
            with open(os.path.join(ngs_dir, study_file)) as segment_file:
                segment_lines = segment_file.readlines()
                segment_lines[0] = 'ID	chrom	loc.start	loc.end	num.mark	seg.mean\n'

            # Write a copy with replaced header
            with open(os.path.join(output_dir, output_file), 'w') as segment_file:
                segment_file.writelines(segment_lines)

            # Create meta file
            meta_filename = os.path.join(output_dir, 'meta_cna_segments.txt')
            create_meta_content(meta_filename, cancer_study_identifier=STUDY_ID,
                                genetic_alteration_type='COPY_NUMBER_ALTERATION', datatype='SEG',
                                reference_genome_id='hg38', description='Segment data', data_filename=output_file)

        # CNA Continuous
        elif 'data_by_genes' in study_file:
            logger.debug('Transforming continuous CNA data: %s' % study_file)

            output_file = 'data_cna_continuous.txt'

            # Read file
            cna_data = pd.read_csv(os.path.join(ngs_dir, study_file), sep='\t', na_values=[''], dtype={'Gene ID': str})
            # Remove column and rename column names
            cna_data.drop('Cytoband', axis=1, inplace=True)
            cna_data.rename(columns={'Gene Symbol': 'Hugo_Symbol', 'Gene ID': 'Entrez_Gene_Id'}, inplace=True)

            # Remove negative Entrez IDs. This can lead to incorrect mapping in cBioPortal
            for index, row in cna_data.iterrows():
                if int(row['Entrez_Gene_Id']) < -1:
                    cna_data.loc[index, 'Entrez_Gene_Id'] = ''
            cna_data.to_csv(os.path.join(output_dir, output_file), sep='\t', index=False, header=True)

            # Create meta file
            meta_filename = os.path.join(output_dir, 'meta_cna_continuous.txt')
            create_meta_content(meta_filename, cancer_study_identifier=STUDY_ID,
                                genetic_alteration_type='COPY_NUMBER_ALTERATION', datatype='LOG2-VALUE',
                                stable_id='log2CNA', show_profile_in_analysis_tab='false',
                                profile_name='Copy-number alteration values',
                                profile_description='Continuous copy-number alteration values for each gene.',
                                data_filename=output_file)

            # Create case list
            cna_samples = cna_data.columns[2:].tolist()
            create_caselist(output_dir=output_dir, file_name='cases_cna.txt', cancer_study_identifier=STUDY_ID,
                            stable_id='%s_cna' % STUDY_ID, case_list_name='CNA samples',
                            case_list_description='All CNA samples', case_list_category='all_cases_with_cna_data',
                            case_list_ids="\t".join(cna_samples))

            # Test for samples in CNA files that are not in clinical data
            if not set(clinical_sample_ids).issuperset(set(cna_samples)):
                logger.error("Found samples in CNA files that are not in clinical data: {}".format(
                    ", ".join(set(cna_samples).difference(set(clinical_sample_ids))))
                )
                raise DataException('Found samples in CNA files that are not in clinical data')

        # CNA Discrete
        elif 'thresholded.by_genes' in study_file:
            logger.debug('Transforming discrete CNA data: %s' % study_file)

            output_file = 'data_cna_discrete.txt'

            # Read input file
            cna_data = pd.read_csv(os.path.join(ngs_dir, study_file), sep='\t', na_values=[''], dtype={'Gene ID': str})
            # Remove column and rename column names
            cna_data.drop('Cytoband', axis=1, inplace=True)
            cna_data.rename(columns={'Gene Symbol': 'Hugo_Symbol', 'Locus ID': 'Entrez_Gene_Id'}, inplace=True)

            # Remove negative Entrez IDs. This can lead to incorrect mapping in cBioPortal
            for index, row in cna_data.iterrows():
                if int(row['Entrez_Gene_Id']) < -1:
                    cna_data.loc[index, 'Entrez_Gene_Id'] = ''
            cna_data.to_csv(os.path.join(output_dir, output_file), sep='\t', index=False, header=True)

            # Create meta file
            meta_filename = os.path.join(output_dir, 'meta_cna_discrete.txt')
            profile_description = 'Putative copy-number alteration values for each gene from GISTIC 2.0.' \
                                  'Values: -2 = homozygous deletion; -1 = hemizygous deletion;' \
                                  '0 = neutral / no change; 1 = gain; 2 = high level amplification.'

            create_meta_content(meta_filename, cancer_study_identifier=STUDY_ID,
                                genetic_alteration_type='COPY_NUMBER_ALTERATION', datatype='DISCRETE',
                                stable_id='gistic', show_profile_in_analysis_tab='true',
                                profile_name='Putative copy-number alterations from GISTIC',
                                profile_description=profile_description,
                                data_filename=output_file)
        elif study_file.split('.')[-2:] == ['maf', 'gz']:
            # Mutations file are transformed in an other loop
            pass

        else:
            logger.warning("Unknown file type: %s" % study_file)
    return cna_samples


def create_cbioportal_study(input_dir: str, ngs_dir: Optional[str], output_dir: str):
    prepare_output_directory(output_dir)

    logger.info('Reading clinical data: %s' % input_dir)
    clinical_sample_ids = process_clinical_data(input_dir, output_dir)

    if ngs_dir:
        logger.info('Reading NGS data: %s' % ngs_dir)
        mutation_samples = process_mutation_data(ngs_dir, output_dir, clinical_sample_ids)
        cna_samples = process_cna_files(ngs_dir, output_dir, clinical_sample_ids)

        # Create cnaseq case list
        cnaseq_samples = list(mutation_samples.union(cna_samples))
        if len(cnaseq_samples) > 0:
            create_caselist(output_dir=output_dir, file_name='cases_cnaseq.txt', cancer_study_identifier=STUDY_ID,
                            stable_id='%s_cnaseq' % STUDY_ID, case_list_name='Sequenced and CNA samples',
                            case_list_description='All sequenced and CNA samples',
                            case_list_category='all_cases_with_mutation_and_cna_data',
                            case_list_ids="\t".join(cnaseq_samples))

    # Create meta study file
    meta_filename = os.path.join(output_dir, 'meta_study.txt')
    create_meta_content(meta_filename, STUDY_ID, type_of_cancer=TYPE_OF_CANCER, name=NAME, short_name=NAME_SHORT,
                        description=DESCRIPTION, add_global_case_list='true')

    logger.info('Done transforming files for %s' % STUDY_ID)

    # Transformation completed
    logger.info('Transformation of studies complete.')
    return


def combine_maf(ngs_dir, output_file_location):
    """
    combines all found NGS files in one. It filters out variants without hugo symbol.
    :param ngs_dir: directory with NGS files
    :param output_file_location: the result NGS file
    :return: unique list of samples in the result file
    """
    samples = set()

    paths_to_process = get_paths_to_non_hidden_maf_gz_files(ngs_dir)

    if not paths_to_process:
        return samples

    header = get_complete_header(paths_to_process)

    if not header:
        return samples

    with open(output_file_location, 'w') as result_maf:
        writer = csv.DictWriter(result_maf, delimiter='\t', fieldnames=header)
        writer.writeheader()
        for study_file in paths_to_process:
            logger.debug('Processing NGS data file: {}'.format(study_file))
            with gzip.open(study_file, 'rt') as file:
                reader = csv.DictReader(not_commented_lines(file), delimiter='\t')
                for row in reader:
                    samples.add(row['Tumor_Sample_Barcode'])
                    writer.writerow(row)
        return samples


def not_commented_lines(iter):
    for line in iter:
        if not line.lstrip().startswith('#'):
            yield line


def get_paths_to_non_hidden_maf_gz_files(ngs_dir):
    paths_to_process = []
    for study_file in os.listdir(ngs_dir):
        if not study_file.startswith('.') and study_file.split('.')[-2:] == ['maf', 'gz']:
            paths_to_process.append(os.path.join(ngs_dir, study_file))
    return paths_to_process


def get_complete_header(paths_to_process):
    fieldnames = []
    for study_file in paths_to_process:
        with gzip.open(study_file, 'rt') as file:
            header = next(csv.reader(not_commented_lines(file), delimiter='\t'))
            for column in header:
                if column not in fieldnames:
                    fieldnames.append(column)
    return fieldnames


def csr2cbioportal(input_dir: str, ngs_dir: Optional[str], output_dir: str):
    logger.info('csr2cbioportal')
    try:
        create_cbioportal_study(input_dir, ngs_dir, output_dir)
    except Exception as e:
        logger.error(e)
        sys.exit(1)


@click.command()
@click.argument('input_dir', type=click.Path(file_okay=False, exists=True, readable=True))
@click.argument('output_dir', type=click.Path(file_okay=False, writable=True))
@click.option('--ngs-dir', type=click.Path(file_okay=False, exists=True, readable=True))
@click.option('--debug', is_flag=True, help='Print more verbose messages')
@click.version_option()
def run(input_dir, ngs_dir, output_dir, debug: bool):
    setup_logging(debug)
    csr2cbioportal(input_dir, ngs_dir, output_dir)


def main():
    run()


if __name__ == '__main__':
    main()



import os
from typing import Optional, Sequence, Dict, List

from csr.exceptions import ReaderException
from sources2csr.ngs import AnalysisType, NGS, LibraryStrategy


class NgsReader:
    """Reader that reads NGS data files.
    """
    def __init__(self, input_dir: str, library_strategy: LibraryStrategy):
        self.input_dir = input_dir
        self.library_strategy = library_strategy

    @staticmethod
    def determine_analysis_type(filename: str) -> Optional[AnalysisType]:
        """Determine whole genome or whole exome sequencing analysis type based on the filename.
        If _WGS or _WXS are not in the filename returns None.

        :param filename: name of the input file
        :return: Type of sequencing event
        """
        if '_WGS' in filename.upper():
            return AnalysisType.WGS
        elif '_WXS' in filename.upper():
            return AnalysisType.WXS
        else:
            return None

    @staticmethod
    def biosource_biomaterial_from_sample_id(sample_id: str, filename: str, by='_') -> tuple:
        """ Splits sample _id into biosource_id and biomaterial_id.
        Every sample ID should be structured as <BiosourceID>_<BiomaterialID>,
        e.g. PMCBS000AAA_PMCBM000AAA.

        :param sample_id: cBioPortal sample ID
        :param filename: ull path with a name of the input file
        :param by: Character to use for splitting, default: '_'
        :return: biosource_id and biomaterial_id values tuple or NgsReaderException
        """

        biosource_biomaterial_pair = sample_id.split(by)
        if len(biosource_biomaterial_pair) != 2:
            raise ReaderException('Invalid sample_id format found in {} NGS file. sample_id: {}'
                                  .format(filename, sample_id))
        else:
            return biosource_biomaterial_pair[0], biosource_biomaterial_pair[1]

    @staticmethod
    def list_files(input_dir: str) -> List[str]:
        """ Get list of file names inside a directory
        :param input_dir: directory with input files
        :return: List of file paths with names
        """
        for file in os.listdir(input_dir):
            if os.path.isfile(os.path.join(input_dir, file)):
                yield file

    def map_ngs(self, biosources_biomaterials_dict: Dict[str, Sequence[str]], filename: str) -> Optional[Sequence[NGS]]:
        """ Maps biosource to biomaterials dictionary to the NGS Sequence

        :param biosources_biomaterials_dict: biosource to biomaterials dictionary
        :param filename: name of the input file
        :return:
        """
        ngs_data = []
        analysis_type = self.determine_analysis_type(filename)
        for biosource, biomaterials in biosources_biomaterials_dict.items():
            for biomaterial in biomaterials:
                ngs_data.append(NGS(biosource_id=biosource,
                                    biomaterial_id=biomaterial,
                                    analysis_type=analysis_type,
                                    library_strategy=self.library_strategy))
        return ngs_data

    def read_data(self, filename: str) -> Optional[Sequence[NGS]]:
        pass



#!/usr/bin/env python3

# Code to create meta files
# Author: Sander Tan, The Hyve

import sys
import logging

logger = logging.getLogger(__name__)
logger.name = logger.name.rsplit('.', 1)[1]
sys.dont_write_bytecode = True


def create_meta_content(file_name,
                        cancer_study_identifier=None,
                        genetic_alteration_type=None,
                        datatype=None,
                        data_filename=None,
                        stable_id=None,
                        profile_name=None,
                        profile_description=None,
                        show_profile_in_analysis_tab=None,
                        type_of_cancer=None,
                        name=None,
                        short_name=None,
                        description=None,
                        add_global_case_list=None,
                        variant_classification_filter=None,
                        pmid=None,
                        reference_genome_id=None,
                        swissprot_identifier=None
                        ):

    # Required properties
    meta_content = []
    if cancer_study_identifier is not None:
        meta_content.append('cancer_study_identifier: {}'.format(cancer_study_identifier))

    # Properties that are required for most data types
    if genetic_alteration_type is not None:
        meta_content.append('genetic_alteration_type: {}'.format(genetic_alteration_type))
        meta_content.append('datatype: {}'.format(datatype))
        meta_content.append('data_filename: {}'.format(data_filename))

    # Properties that are required for mutation, gene expression and protein enrichment profiles
    if stable_id is not None:
        meta_content.append('stable_id: {}'.format(stable_id))
        meta_content.append('profile_name: {}'.format(profile_name))
        meta_content.append('profile_description: {}'.format(profile_description))
        meta_content.append('show_profile_in_analysis_tab: {}'.format(show_profile_in_analysis_tab))

    # Properties that are required for meta_study
    if name is not None:
        meta_content.append('type_of_cancer: {}'.format(type_of_cancer))
        meta_content.append('name: {}'.format(name))
        meta_content.append('short_name: {}'.format(short_name))
        meta_content.append('description: {}'.format(description))
        meta_content.append('add_global_case_list: {}'.format(add_global_case_list))

    if variant_classification_filter is not None:
        meta_content.append('variant_classification_filter: {}'.format(variant_classification_filter))

    if swissprot_identifier is not None:
        meta_content.append('swissprot_identifier: {}'.format(swissprot_identifier))

    if pmid is not None:
        meta_content.append('pmid: {}'.format(pmid))

    if reference_genome_id is not None:
        meta_content.append('reference_genome_id: {}'.format(reference_genome_id))
        meta_content.append('description: {}'.format(description))

    # Write file
    logger.debug('Writing cBioPortal metadata to {} for study {} and datatype {}'
                 .format(file_name, cancer_study_identifier, datatype))
    with open(file_name, 'w') as meta_output_file:
        meta_output_file.write('\n'.join(meta_content) + '\n')
    return



import json
import logging
from datetime import datetime
from typing import List, Dict, Any, Type
from pydantic import BaseModel
from csr.tabular_file_reader import TabularFileReader

logger = logging.getLogger(__name__)


class EntityReader:
    """Reader that reads entity data from tab delimited files.
    """
    def __init__(self, input_dir: str):
        self.input_dir = input_dir

    @staticmethod
    def get_date_fields(schema: Dict[str, Any]) -> List[str]:
        date_fields: List[str] = []
        for field in schema['properties'].values():
            if field.get('format') == 'date':
                date_fields.append(field['title'].lower())
        return date_fields

    @staticmethod
    def get_array_fields(schema: Dict[str, Any]) -> List[str]:
        return [field['title'].lower()
                for field in schema['properties'].values()
                if field.get('type') == 'array']

    def read_entities(self, file_path: str, entity_type: Type[BaseModel]) -> List[Any]:
        try:
            data = TabularFileReader(file_path).read_data()
        except FileNotFoundError:
            return []

        date_fields = self.get_date_fields(entity_type.schema())
        array_fields = self.get_array_fields(entity_type.schema())

        for row in data:
            for field, value in row.items():
                if value == '' or value == 'NA':
                    row[field] = None
                elif field in date_fields:
                    row[field] = datetime.strptime(value, '%Y-%m-%d')
                elif field in array_fields:
                    row[field] = json.loads(value)
        return list(data)



import os
from typing import Optional, Sequence

from csr.tabular_file_reader import TabularFileReader
from sources2csr.ngs import NGS, LibraryStrategy
from sources2csr.ngs_reader import NgsReader, ReaderException


class NgsTxtReader(NgsReader):
    """ Reads .txt files with Continuous CNA per gene and Discrete CNA per gene data.
    """

    def __init__(self, input_dir: str):
        super().__init__(input_dir, LibraryStrategy.CNV)

    def read_data(self, filename: str) -> Optional[Sequence[NGS]]:
        """ Reads .txt file.
        Sample_id should be specified in the header. Assumes that the IDs will start with 'PMC'.

        :param filename: name of the input file
        :return: Sequence of NGS objects
        """
        data = TabularFileReader(os.path.join(self.input_dir, filename)).read_data()
        biosource_biomaterial_dict = dict()
        if data:
            sample_id_col_num = 0
            for col_value in data[0]:
                if col_value.startswith('PMC'):
                    sample_id_col_num += 1
                    biosource_biomaterial = self.biosource_biomaterial_from_sample_id(col_value, filename)
                    biosource_biomaterial_dict.setdefault(biosource_biomaterial[0], []).append(biosource_biomaterial[1])
            if sample_id_col_num == 0:
                raise ReaderException("Cannot read NGS data from file: {}. No sample_id found in header"
                                      .format(filename))
        else:
            raise ReaderException("Cannot read NGS data from file: {}. Empty data.".format(filename))
        return self.map_ngs(biosource_biomaterial_dict, filename)



import logging
import sys

import click

from csr.logging import setup_logging
from csr.subject_registry_writer import SubjectRegistryWriter

from csr.study_registry_writer import StudyRegistryWriter
from sources2csr.derived_values import add_derived_values
from sources2csr.ngs2csr import add_ngs_data
from sources2csr.sources_reader import SourcesReader


logger = logging.getLogger(__name__)


def sources2csr(input_dir, output_dir, config_dir):
    logger.info('sources2csr')
    try:
        reader = SourcesReader(input_dir=input_dir, config_dir=config_dir)
        subject_registry = reader.read_subject_data()
        add_derived_values(subject_registry)
        add_ngs_data(subject_registry, input_dir)
        subject_registry_writer = SubjectRegistryWriter(output_dir)
        subject_registry_writer.write(subject_registry)

        study_registry = reader.read_study_data()
        study_registry_writer = StudyRegistryWriter(output_dir)
        study_registry_writer.write(study_registry)
    except Exception as e:
        logger.error(e)
        sys.exit(1)


@click.command()
@click.argument('input_dir', type=click.Path(file_okay=False, exists=True, readable=True))
@click.argument('output_dir', type=click.Path(file_okay=False, writable=True))
@click.argument('config_dir', type=click.Path(file_okay=False, exists=True, readable=True))
@click.option('--debug', is_flag=True, help='Print more verbose messages')
@click.version_option()
def run(input_dir, output_dir, config_dir, debug: bool):
    setup_logging(debug)
    sources2csr(input_dir, output_dir, config_dir)


def main():
    run()


if __name__ == '__main__':
    main()



import json
import logging
import sys
from os import path

import click

from csr.exceptions import DataException
from csr.logging import setup_logging
from transmart_loader.copy_writer import TransmartCopyWriter
from transmart_loader.transmart import DataCollection

from csr.csr import CentralSubjectRegistry, StudyRegistry
from csr.subject_registry_reader import SubjectRegistryReader
from csr.study_registry_reader import StudyRegistryReader
from csr2transmart.mappers.csr_mapper import CsrMapper
from csr2transmart.ontology_config import OntologyConfig

logger = logging.getLogger(__name__)


def read_configuration(config_dir) -> OntologyConfig:
    """ Parse configuration files and return set of dictionaries

    :param config_dir: Path to directory where the configs are stored
    """
    ontology_config_path = path.join(config_dir, 'ontology_config.json')
    if not path.exists(ontology_config_path) or not path.isfile(ontology_config_path):
        raise FileNotFoundError(f'Cannot find {ontology_config_path}')
    with open(ontology_config_path, 'r') as ontology_config_file:
        try:
            config_data = json.load(ontology_config_file)
        except Exception as e:
            logger.error(e)
            raise DataException(f'Error parsing ontology config file: {ontology_config_path}')
        return OntologyConfig(**config_data)


def csr2transmart(input_dir: str,
                  output_dir: str,
                  config_dir: str,
                  study_id: str,
                  top_tree_node: str):
    logger.info('csr2transmart')
    try:
        logger.info('Reading configuration data...')
        ontology_config = read_configuration(config_dir)

        logger.info('Reading CSR data...')
        subject_registry_reader = SubjectRegistryReader(input_dir)
        subject_registry: CentralSubjectRegistry = subject_registry_reader.read_subject_registry()
        study_registry_reader = StudyRegistryReader(input_dir)
        study_registry: StudyRegistry = study_registry_reader.read_subject_registry()

        logger.info('Mapping CSR to Data Collection...')
        mapper = CsrMapper(study_id, top_tree_node)
        collection: DataCollection = mapper.map(subject_registry, study_registry, ontology_config.nodes)

        logger.info('Writing files to {}'.format(output_dir))
        copy_writer = TransmartCopyWriter(str(output_dir))
        copy_writer.write_collection(collection)

        logger.info('Done.')

    except Exception as e:
        logger.error(e)
        sys.exit(1)


@click.command()
@click.argument('input_dir', type=click.Path(file_okay=False, exists=True, readable=True))
@click.argument('output_dir', type=click.Path(file_okay=False, writable=True))
@click.argument('config_dir', type=click.Path(file_okay=False, exists=True, readable=True))
@click.option('--debug', is_flag=True, help='Print more verbose messages')
@click.version_option()
def run(input_dir, output_dir, config_dir, debug: bool):
    setup_logging(debug)
    csr2transmart(
        input_dir,
        output_dir,
        config_dir,
        'CSR',
        '\\Central Subject Registry\\',
    )


def main():
    run()


if __name__ == '__main__':
    main()



from collections import Counter
from typing import Dict, Sequence, Optional

from pydantic import BaseModel, Schema, validator


class SourcesConfigValidationException(ValueError):
    pass


class Source(BaseModel):
    """
    The source file column for the entity attribute.
    If the column is not specified, it is assumed that the column name
    is the attribute name.
    """
    file: str
    column: Optional[str]
    date_format: Optional[str]


class Attribute(BaseModel):
    name: str
    sources: Sequence[Source]


class Entity(BaseModel):
    attributes: Sequence[Attribute]

    @validator('attributes', whole=True)
    def validate_attribute_names(cls, attrs: Sequence[Attribute]):
        attribute_name_counts = Counter([attr.name for attr in attrs])
        duplicates = [k for k, v in attribute_name_counts.items() if v > 1]
        if duplicates:
            raise SourcesConfigValidationException(
                f'Duplicate attributes: {", ".join(duplicates)}')
        return attrs


class FileFormat(BaseModel):
    delimiter: str = Schema('\t', min_length=1, max_length=1)


class SourcesConfig(BaseModel):

    """Mapping from entity name to entity sources"""
    entities: Dict[str, Entity]

    """Associates input files with codebook files"""
    codebooks: Optional[Dict[str, str]]

    """File format configuration for input files"""
    file_format: Optional[Dict[str, FileFormat]]



from collections import Counter
from typing import Sequence, Optional

from csr.csr import SubjectEntity, StudyEntity
from pydantic import BaseModel, validator, constr


class OntologyConfigValidationException(ValueError):
    pass


class TreeNode(BaseModel):
    """
    Concept node
    """
    name: constr(min_length=1)
    concept_code: Optional[str]
    children: Optional[Sequence['TreeNode']]

    @validator('children', always=True)
    def check_consistency(cls, v, values):
        if v is not None and values.get('concept_code') is not None:
            raise OntologyConfigValidationException(f'Node cannot have both concept_code and children: {values}')
        if v is None and values.get('concept_code') is None:
            raise OntologyConfigValidationException(f'Node must have either concept_code or children: {values}')
        return v

    @validator('concept_code')
    def check_valid_in_csr(cls, concept_code):
        entity_name_field_pair = concept_code.split('.')
        if len(entity_name_field_pair) != 2:
            raise OntologyConfigValidationException(f'Invalid concept code format: {concept_code}. '
                                                    f'Concept code has to have format: `<entity_name>.<entity_field>`')

        csr_entities = list(SubjectEntity.__args__) + list(StudyEntity.__args__)
        csr_entity_names = list(map(lambda se: se.schema()['title'], csr_entities))
        if entity_name_field_pair[0] not in csr_entity_names:
            raise OntologyConfigValidationException(f'Invalid concept code: {concept_code}. '
                                                    f'{entity_name_field_pair[0]} is not a valid CSR entity.')

        csr_entity_fields = list(filter(lambda e: e.schema()['title'] == entity_name_field_pair[0],
                                        csr_entities))[0].schema()['properties'].keys()
        if entity_name_field_pair[1] not in csr_entity_fields:
            raise OntologyConfigValidationException(f'Invalid concept code: {concept_code}. '
                                                    f'Field: {entity_name_field_pair[1]} '
                                                    f'is not a valid field of {entity_name_field_pair[0]} entity.')
        return concept_code

    @validator('children', whole=True)
    def validate_child_names(cls, nodes: Optional[Sequence['TreeNode']]):
        if nodes is None:
            return nodes
        node_name_counts = Counter([node.name for node in nodes])
        duplicates = [k for k, v in node_name_counts.items() if v > 1]
        if duplicates:
            raise OntologyConfigValidationException(
                f'Duplicate child names: {", ".join(duplicates)}')
        return nodes


TreeNode.update_forward_refs()


class OntologyConfig(BaseModel):
    """
    Ontology tree configuration
    """
    nodes: Sequence[TreeNode]

    @validator('nodes', whole=True)
    def validate_node_names(cls, nodes: Sequence[TreeNode]):
        node_name_counts = Counter([node.name for node in nodes])
        duplicates = [k for k, v in node_name_counts.items() if v > 1]
        if duplicates:
            raise OntologyConfigValidationException(
                f'Duplicate node names: {", ".join(duplicates)}')
        return nodes



import logging
from os import path

from csr.csr import CentralSubjectRegistry, Individual, Diagnosis, Biosource, Biomaterial
from csr.entity_reader import EntityReader

logger = logging.getLogger(__name__)


class SubjectRegistryReader(EntityReader):
    """Reader that reads Central Subject Registry (CSR) data from tab delimited files.
    """

    def __init__(self, input_dir: str):
        EntityReader.__init__(self, input_dir)

    def read_subject_registry(self) -> CentralSubjectRegistry:
        try:
            individuals = self.read_entities(path.join(self.input_dir, 'individual.tsv'), Individual)
            diagnoses = self.read_entities(path.join(self.input_dir, 'diagnosis.tsv'), Diagnosis)
            biosources = self.read_entities(path.join(self.input_dir, 'biosource.tsv'), Biosource)
            biomaterials = self.read_entities(path.join(self.input_dir, 'biomaterial.tsv'), Biomaterial)
            return CentralSubjectRegistry(individuals=individuals,
                                          diagnoses=diagnoses,
                                          biosources=biosources,
                                          biomaterials=biomaterials)
        except FileNotFoundError as fnfe:
            raise FileNotFoundError('File not found. {}'.format(fnfe))



import logging

from csr.csr import StudyRegistry, Study, IndividualStudy
from csr.entity_reader import EntityReader

logger = logging.getLogger(__name__)


class StudyRegistryReader(EntityReader):
    """Reader that reads study registry data from tab delimited files.
    """

    def __init__(self, input_dir: str):
        EntityReader.__init__(self, input_dir)

    def read_subject_registry(self) -> StudyRegistry:
        try:
            studies = self.read_entities('{}/{}'.format(self.input_dir, 'study.tsv'), Study)
            individual_studies = self.read_entities('{}/{}'.format(self.input_dir, 'individual_study.tsv'),
                                                    IndividualStudy)
            return StudyRegistry(studies=studies,
                                 individual_studies=individual_studies)
        except FileNotFoundError as fnfe:
            raise FileNotFoundError('File not found. {}'.format(fnfe))



import csv
import gzip
import os
from typing import Sequence, Dict, Any

from csr.exceptions import ReaderException


class TabularFileReader:
    """
    Delimiter-separated values reader.
    """
    def __iter__(self):
        return self.reader.__iter__()

    def read_data(self) -> Sequence[Dict[str, Any]]:
        data = []
        first = True
        header = None
        line_num = 0
        for line in self:
            line_num += 1
            if line and str.startswith(line[0], '#'):
                # skip comment lines
                continue
            if first:
                header = line
                first = False
            else:
                if not len(line) == len(header):
                    raise ReaderException(f'Unexpected line length {len(line)}. Expected {len(header)}. '
                                          f'File {self.path}, line number: {line_num}.')
                record = dict([(header[i], line[i]) for i in range(0, len(header))])
                data.append(record)
        return data

    def close(self) -> None:
        if self.file:
            self.file.close()

    def __init__(self, path: str, delimiter='\t'):
        self.file = None
        self.path = path
        if not os.path.isfile(path):
            raise ReaderException(f'File not found: {path}')
        if str.endswith(path, '.gz'):
            self.file = gzip.open(path, 'rt')
        else:
            self.file = open(path, 'r')
        self.reader = csv.reader(self.file, delimiter=delimiter)

    def __del__(self):
        self.close()



import logging.config
import os

import yaml


def setup_logging(debug: bool):
    """
    Setup logging configuration
    :param debug: flag if debug messages should be displayed
    :param name: application name to write the log for
    """
    default_level = logging.DEBUG if debug else logging.INFO
    path = os.environ.get('LOG_CFG', 'csr/logging.yaml')
    if os.path.exists(path):
        with open(path, 'rt') as f:
            config = yaml.safe_load(f.read())
        config['root']['level'] = default_level
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)



from typing import List, Dict, Optional, Sequence, Type
from pydantic import BaseModel
from transmart_loader.transmart import TrialVisit, Patient, Concept, Modifier, Observation, ObservationMetadata, \
    Value, CategoricalValue, ValueType, NumericalValue, DateValue, TextValue

from csr.csr import CentralSubjectRegistry, StudyRegistry, Individual, SubjectEntity

from csr.exceptions import MappingException


class ObservationMapper:
    """
    Map observations for subject registry and study registry
    """
    def __init__(self,
                 default_trial_visit: TrialVisit,
                 individual_id_to_patient: Dict[str, Patient],
                 concept_code_to_concept: Dict[str, Concept],
                 modifier_key_to_modifier: Dict[str, Modifier]):
        self.default_trial_visit = default_trial_visit
        self.individual_id_to_patient = individual_id_to_patient
        self.concept_code_to_concept = concept_code_to_concept
        self.modifier_key_to_modifier = modifier_key_to_modifier
        self.observations: List[Observation] = []

    @staticmethod
    def row_value_to_value(row_value, value_type: ValueType) -> Optional[Value]:
        """
        Map entity value to transmart-loader Value by ValueType
        :param row_value: entity value
        :param value_type: transmart-loader ValueType
        :return: transmart-loader Value
        """
        if row_value is None:
            return None
        if value_type is ValueType.Categorical:
            return CategoricalValue(row_value)
        elif value_type is ValueType.Numeric:
            return NumericalValue(row_value)
        elif value_type is ValueType.DateValue:
            return DateValue(row_value)
        else:
            return TextValue(row_value)

    @staticmethod
    def get_field_name_by_keyword(entity_type: Type[BaseModel], key: str, value) -> str:
        """
        Get name of the field of CSR entity type by the schema metadata
        :param entity_type: type of the CSR entity
        :param key: schema keyword key
        :param value: schema keyword value
        :return:
        """
        return list([name for (name, prop) in entity_type.schema()['properties'].items()
                     if key in prop and prop[key] is value])[0]

    def get_id_field_name(self, entity_type: Type[BaseModel]) -> str:
        """
        Get identifying field name of a CSR entity type by 'identity' schema keyword
        :param entity_type: type of the CSR entity
        :return:
        """
        return self.get_field_name_by_keyword(entity_type, 'identity', True)

    def get_ref_field_name(self, entity_type: Type[BaseModel], ref_type: str) -> str:
        """
        Get a name of the field referencing to entity of specified type
        :param entity_type: type of the CSR entity
        :param ref_type: name of the referencing entity type
        :return:
        """
        return self.get_field_name_by_keyword(entity_type, 'references', ref_type)

    def map_observation_metadata(self, modifier_key, value: str) -> Optional[ObservationMetadata]:
        """
        Get observation modifier
        :param modifier_key: observation modifier key
        :param value: value of the modifier observation
        :return: transmart-loader metadata if any
        """
        if modifier_key is None or value is None:
            return None
        mod_metadata: Dict[Modifier, Value] = dict()
        modifier = self.modifier_key_to_modifier.get(modifier_key)
        if modifier is None:
            return None
        mod_metadata[modifier] = CategoricalValue(value)
        return ObservationMetadata(mod_metadata)

    def get_observation_for_value(self, row_value, concept: Concept, metadata: ObservationMetadata,
                                  patient: Patient) -> Observation:
        value = self.row_value_to_value(row_value, concept.value_type)
        return Observation(patient, concept, None, self.default_trial_visit, None, None, value, metadata)

    def map_observation(self, entity: BaseModel, entity_id: str, patient: Patient):
        """
        Map entity to trasmart-loader Observation
        :param entity: CSR entity
        :param entity_id: id of the entity
        :param patient: individual linked to the observation
        :return:
        """
        entity_fields = entity.fields.keys()
        entity_name = entity.schema()['title']
        for entity_field in entity_fields:
            concept_code = '{}.{}'.format(entity_name, entity_field)
            concept = self.concept_code_to_concept.get(concept_code)
            if concept is not None:
                if isinstance(entity, Individual):
                    metadata = None
                else:
                    metadata = self.map_observation_metadata(entity_name, entity_id)
                entity_value = getattr(entity, entity_field)
                if entity_value is not None:
                    if isinstance(entity_value, List):
                        for v in entity_value:
                            self.observations.append(
                                self.get_observation_for_value(v, concept, metadata, patient))
                    else:
                        self.observations.append(
                            self.get_observation_for_value(entity_value, concept, metadata, patient))

    def map_individual_linked_entity_observations(self, entities: Sequence[BaseModel]):
        """
        Map observations for entities that have a direct link to individuals,
        :param entities: Central subject registry entities
        :return:
        """
        if entities:
            entity_type = type(entities[0])
            id_attribute = self.get_id_field_name(entity_type)
            for entity in entities:
                entity_id = getattr(entity, id_attribute)
                patient = self.individual_id_to_patient.get(entity.individual_id)
                if patient is None:
                    raise MappingException('No patient with identifier: {}. '
                                           'Failed to create observation for {} with id: {}. Entity {}, Ind {}'
                                           .format(entity.individual_id, type(entity).__name__, entity_id,
                                                   entity, self.individual_id_to_patient.keys()))
                self.map_observation(entity, entity_id, patient)

    def map_non_individual_linked_entity_observations(self,
                                                      entities: Sequence[SubjectEntity],
                                                      ref_entities: Sequence[SubjectEntity]):
        """
        Map observations for entities that do not have a direct link to individuals,
        but have a reference linking to individuals.
        :param entities: Central subject registry entities
        :param ref_entities: reference entities that have a link to individuals
        :return:
        """
        if entities and ref_entities:
            entity_type = type(entities[0])
            ref_entity_type = type(ref_entities[0])
            entity_id_field_name = self.get_id_field_name(entity_type)
            ref_id_field_name = self.get_id_field_name(ref_entity_type)
            entity_ref_field_name = self.get_ref_field_name(entity_type, ref_entity_type.schema()['title'])
            for entity in entities:
                entity_id = entity.__getattribute__(entity_id_field_name)
                linked_entity = next((
                    re for re in ref_entities
                    if entity.__getattribute__(entity_ref_field_name) == re.__getattribute__(ref_id_field_name)), None)
                if linked_entity is None:
                    raise MappingException('No {} linked to {} with id: {}. '
                                           'Failed to create observation.'
                                           .format(ref_entity_type.schema()['title'],
                                                   entity_type.schema()['title'],
                                                   entity_id))
                patient = self.individual_id_to_patient.get(linked_entity.individual_id)
                if patient is None:
                    raise MappingException('No patient with identifier: {}. '
                                           'Failed to create observation for {} with id: {}.'
                                           .format(linked_entity.individual_id, entity.schema()['title'],
                                                   entity_id))
                self.map_observation(entity, entity_id, patient)

    def map_study_registry_observations(self, study_registry: StudyRegistry):
        """
        Map observations for Subject registry entities
        :param study_registry: Study registry
        :return:
        """
        for ind_study in study_registry.individual_studies:
            study = next((s for s in study_registry.studies if s.study_id == ind_study.study_id), None)
            if study is None:
                raise MappingException('No study with identifier: {}. '
                                       'Failed to create observation for individual study with id: {}.'
                                       .format(ind_study.study_id, ind_study.individual_id))
            patient = self.individual_id_to_patient.get(ind_study.individual_id)
            if patient is None:
                raise MappingException('No patient with identifier: {}. '
                                       'Skipping creating observation for study with id: {}.'
                                       .format(ind_study.individual_id, ind_study.individual_id))
            self.map_observation(study, study.study_id, patient)

    def map_subject_registry_observations(self,
                                          entities: Sequence[SubjectEntity],
                                          ref_entities: Sequence[SubjectEntity] = []):
        """
        Map observations for Central subject registry entities
        :param entities: Central subject registry entities
        :param ref_entities: Optional, Central subject registry entities that link to individuals
        :return:
        """
        if not ref_entities:
            self.map_individual_linked_entity_observations(entities)
        else:
            self.map_non_individual_linked_entity_observations(entities, ref_entities)

    def map_observations(self, subject_registry: CentralSubjectRegistry, study_registry: StudyRegistry):
        """
        Map observations for each csr entity
        :param subject_registry: Central subject registry
        :param study_registry: Study registry
        :return:
        """
        self.map_subject_registry_observations(subject_registry.individuals)
        self.map_subject_registry_observations(subject_registry.diagnoses)
        self.map_subject_registry_observations(subject_registry.biosources)
        self.map_subject_registry_observations(subject_registry.biomaterials, subject_registry.biosources)
        self.map_study_registry_observations(study_registry)



from collections import defaultdict
from typing import List, Dict
from dateutil.relativedelta import relativedelta

from csr.csr import CentralSubjectRegistry, Diagnosis


def add_derived_values(subject_registry: CentralSubjectRegistry) -> CentralSubjectRegistry:
    """Compute derived diagnosis aggregate values
    :param subject_registry: Central Subject Registry
    :return: updated Central Subject Registry
    """

    # Collect relevant diagnosis aggregates per individual
    diagnoses_per_individual: Dict[str, List[Diagnosis]] = defaultdict(list)
    for diagnosis in subject_registry.diagnoses:
        diagnoses_per_individual[diagnosis.individual_id].append(diagnosis)
    diagnosis_count_per_individual = {}
    first_diagnosis_date_per_individual = {}
    for individual_id, diagnoses in diagnoses_per_individual.items():
        diagnosis_count_per_individual[individual_id] = len(diagnoses)
        diagnosis_dates = [d.diagnosis_date for d in diagnoses if d.diagnosis_date is not None]
        if diagnosis_dates:
            first_diagnosis_date = sorted(diagnosis_dates)[0]
            first_diagnosis_date_per_individual[individual_id] = first_diagnosis_date

    # Add diagnosis aggregates to individuals
    for individual in subject_registry.individuals:
        # Add diagnosis count
        individual.diagnosis_count = diagnosis_count_per_individual.get(individual.individual_id, None)
        # Add age at first diagnosis
        if individual.birth_date is not None:
            first_diagnosis_date = first_diagnosis_date_per_individual.get(individual.individual_id, None)
            if first_diagnosis_date is not None:
                individual.age_first_diagnosis = relativedelta(first_diagnosis_date, individual.birth_date).years

    return subject_registry



#!/usr/bin/env python3

# Code to create case lists
# Author: Sander Tan, The Hyve

import os
import logging

logger = logging.getLogger(__name__)
logger.name = logger.name.rsplit('.', 1)[1]


def create_caselist(output_dir,
                    file_name,
                    cancer_study_identifier=None,
                    stable_id=None,
                    case_list_name=None,
                    case_list_description=None,
                    case_list_category=None,
                    case_list_ids=None
                    ):

    # Define output directory
    case_list_dir = os.path.join(output_dir, 'case_lists')
    if not os.path.exists(case_list_dir):
        os.mkdir(case_list_dir)

    # Create contents
    caselist_content = ['cancer_study_identifier: {}'.format(cancer_study_identifier),
                        'stable_id: {}'.format(stable_id),
                        'case_list_name: {}'.format(case_list_name),
                        'case_list_description: {}'.format(case_list_description),
                        'case_list_category: {}'.format(case_list_category),
                        'case_list_ids: {}'.format(case_list_ids)
                        ]

    # Write file
    logger.debug('Writing cBioPortal caselist to {} for study {}'.format(file_name, cancer_study_identifier))
    with open(os.path.join(case_list_dir, file_name), 'w') as caselist_output_file:
        caselist_output_file.write('\n'.join(caselist_content) + '\n')
    return



from typing import Any, Dict, List, Optional, Sequence

from csr.exceptions import DataException
from sources2csr.codebook import CodeBook, ColumnValueMapping


def read_codebook(codebook_filename: str) -> CodeBook:
    """Process the content of a codebook and return the reformatted codebook as an object.
    Expected import format is a tab-delimited file.

    Format:
    - A header line is followed by one or more value mapping lines.
    - Header lines contain two columns: a number and a space separated list of column names. E.g., "1\tSEX GENDER".
    - Value mapping lines start with an empty field, followed by code, value pairs. E.g., "\t1\tMale\t2\tFemale".

    :param codebook_filename: file name of the code book
    :return: code book object
    """
    with open(codebook_filename, 'r') as code_book_file:
        column_mappings: Dict[str, ColumnValueMapping] = {}
        current_value_mapping: Dict[Any, Any] = {}
        current_columns: Optional[List[str]] = None
        line_number = 0
        for line in code_book_file:
            line_number += 1
            line = line.strip('\n')
            if not line.startswith('\t'):
                # Save previous column mapping
                if current_columns is not None:
                    column_value_mapping = ColumnValueMapping(value_mapping=current_value_mapping)
                    for column in current_columns:
                        column_mappings[column] = column_value_mapping
                # Start new column mapping
                tokens = line.split('\t')
                if len(tokens) < 2:
                    raise DataException(
                        f'Invalid header in codebook {codebook_filename} on line {line_number}')
                current_columns = [column_name.lower() for column_name in tokens[1].split(' ')]
                duplicate_columns = set(current_columns).intersection(column_mappings.keys())
                if duplicate_columns:
                    raise DataException(
                        f'Duplicate columns in codebook on line {line_number}: {", ".join(duplicate_columns)}')
                current_value_mapping = {}
            else:
                # Add values to current value mapping
                tokens = line.split('\t')[1:]
                if len(tokens) % 2 != 0:
                    raise DataException(
                        f'Invalid value mapping in codebook {codebook_filename} on line {line_number}')
                it = iter(tokens)
                for code, value in zip(it, it):
                    if code != '' and value != '':
                        value = value.replace('"', '')
                        if code in current_value_mapping:
                            raise DataException(f'Duplicate code in codebook on line {line_number}: {code}')
                        current_value_mapping[code] = value
        # Save last column mapping
        if current_columns is not None:
            column_value_mapping = ColumnValueMapping(value_mapping=current_value_mapping)
            for column in current_columns:
                column_mappings[column] = column_value_mapping
        return CodeBook(column_mappings=column_mappings)


def apply_codebook_mapping(mapping: Dict[str, ColumnValueMapping], column: str, value: Any) -> Any:
    column_mapping = mapping.get(column.lower(), None)
    if column_mapping is None:
        return value
    return column_mapping.value_mapping.get(value, value)


class CodeBookMapper:

    def __init__(self, codebook_filename: str):
        self.codebook = read_codebook(codebook_filename)

    def apply(self, data: Sequence[Dict[str, Any]]) -> Sequence[Dict[str, Any]]:
        column_mappings = self.codebook.column_mappings
        data = [{k: apply_codebook_mapping(column_mappings, k, v) for k, v in item.items()} for item in data]
        return data



from csr.csr import CentralSubjectRegistry, Individual, Diagnosis, Biosource, Biomaterial
from csr.entity_writer import EntityWriter


class SubjectRegistryWriter(EntityWriter):
    """Writer that writes Central Subject Registry (CSR) data to tab delimited files.
    """

    def __init__(self, output_dir: str):
        EntityWriter.__init__(self, output_dir)

    def write(self, subject_registry: CentralSubjectRegistry):
        self.write_entities('individual.tsv', Individual.schema(), subject_registry.individuals)
        self.write_entities('diagnosis.tsv', Diagnosis.schema(), subject_registry.diagnoses)
        self.write_entities('biosource.tsv', Biosource.schema(), subject_registry.biosources)
        self.write_entities('biomaterial.tsv', Biomaterial.schema(), subject_registry.biomaterials)



import json
import logging
import os
from os import path
from typing import Optional, Sequence, Dict, List, Any

from pydantic import BaseModel
from csr.exceptions import FileSystemException
from transmart_loader.tsv_writer import TsvWriter


logger = logging.getLogger(__name__)


class EntityWriter:
    """Writer that writes entity data to tab delimited files.
    """

    def __init__(self, output_dir: str):
        self.output_dir = output_dir
        if not path.exists(output_dir):
            logger.info('Creating output directory: {}'.format(output_dir))
            os.makedirs(output_dir, 0o0700, True)

    @staticmethod
    def format_values(values: List[Any]):
        result = []
        for value in values:
            if type(value) is list:
                result.append(json.dumps(value))
            else:
                result.append(value)
        return result

    def write_entities(self, filename: str, schema: Dict, elements: Optional[Sequence[BaseModel]]):
        output_path = self.output_dir + '/' + filename
        if path.exists(output_path):
            raise FileSystemException('File already exists: {}'.format(output_path))
        writer: TsvWriter = TsvWriter(output_path)
        writer.writerow(list(schema['properties'].keys()))
        if elements:
            for element in elements:
                writer.writerow(self.format_values(list(element.dict().values())))



import json
import logging
from datetime import datetime
from math import isnan
from os import path
from typing import Any, Tuple, Dict, Union, Sequence

from pydantic import BaseModel

from csr.csr import CentralSubjectRegistry, StudyRegistry, SubjectEntity, StudyEntity
from csr.tabular_file_reader import TabularFileReader
from sources2csr.codebook_mapper import CodeBookMapper
from csr.exceptions import DataException
from sources2csr.sources_config import SourcesConfig, Entity

logger = logging.getLogger(__name__)


def format_column(column: Union[str, Tuple[str, str]]):
    if isinstance(column, str):
        return column.lower()
    return column[0].lower() if column[1] == '' else column[1].lower()


def format_value(schema: Dict, column: str, value: Any):
    if column not in schema['properties']:
        return None
    if isinstance(value, float) and isnan(value):
        return None
    return value


def get_id_property(schema: Dict) -> str:
    id_property = list([name for (name, prop) in schema['properties'].items()
                        if 'identity' in prop and prop['identity'] is True])[0]
    logger.debug(f'Id property: {id_property}')
    return id_property


def transform_entity(values: Dict[Any, Any], schema: Dict) -> Dict:
    return {format_column(k): format_value(schema, format_column(k), v)
            for k, v in values.items()}


def transform_entities(entities: Any, schema: Dict, constructor: Any):
    id_property = get_id_property(schema)
    entities = [transform_entity(entity_data, schema) for entity_data in entities]
    result = []
    for entity in entities:
        try:
            result.append(constructor(entity))
        except Exception as e:
            logger.error(e)
            entity_name = schema['title']
            id = entity[id_property]
            raise DataException(f'Invalid data for {entity_name} with id {id}')
    return result


def read_configuration(config_dir) -> SourcesConfig:
    """ Parse configuration files and return set of dictionaries

    :param config_dir: Path to directory where the configs are stored
    """
    sources_config_path = path.join(config_dir, 'sources_config.json')
    if not path.exists(sources_config_path) or not path.isfile(sources_config_path):
        raise DataException(f'Cannot find {sources_config_path}')
    with open(sources_config_path, 'r') as sources_config_file:
        try:
            config_data = json.load(sources_config_file)
        except Exception as e:
            logger.error(e)
            raise DataException(f'Error parsing source config file: {sources_config_path}')
        return SourcesConfig(**config_data)


def get_source_files(entity_sources_config: Entity, id_property: str):
    source_files = set([source.file
                        for attribute in entity_sources_config.attributes
                        for source in attribute.sources])
    source_file_id_mapping = dict([(source.file, source.column if source.column is not None else attribute.name)
                                   for attribute in entity_sources_config.attributes
                                   for source in attribute.sources
                                   if attribute.name == id_property])
    logger.debug(f'Source files: {source_files}')
    logger.debug(f'Source file id mapping: {source_file_id_mapping}')
    source_files_without_id_column = source_files - set(source_file_id_mapping.keys())
    if source_files_without_id_column:
        raise DataException(f'Id column missing in source files: {source_files_without_id_column}')
    return source_files, source_file_id_mapping


def validate_derived_values_not_in_source_config(entity_type: BaseModel, entity_source_config: Entity):
    derived_properties = set([name for (name, prop) in entity_type.schema()['properties'].items()
                              if 'derived' in prop and prop['derived'] is True])
    attribute_names = set([attr.name for attr in entity_source_config.attributes])
    intersection = derived_properties.intersection(attribute_names)
    if intersection:
        raise DataException(f'Derived value fields not allowed in source files: {", ".join(intersection)}')


class SourcesReader:

    def __init__(self, input_dir, config_dir):
        self.input_dir = input_dir
        self.sources_config = read_configuration(config_dir)

    def read_source_file_data(self, source_file) -> Sequence[Dict[str, Any]]:
        file_format = self.sources_config.file_format.get(source_file, None)\
            if self.sources_config.file_format else None
        if file_format is not None:
            reader = TabularFileReader(path.join(self.input_dir, source_file), file_format.delimiter)
        else:
            reader = TabularFileReader(path.join(self.input_dir, source_file))
        source_file_data = reader.read_data()
        if self.sources_config.codebooks is not None:
            codebook_filename = self.sources_config.codebooks.get(source_file, None)
            if codebook_filename is not None:
                codebook_mapper = CodeBookMapper(path.join(self.input_dir, codebook_filename))
                source_file_data = codebook_mapper.apply(source_file_data)
        return source_file_data

    def read_id_property(self, entity_type) -> str:
        entity_sources_config = self.sources_config.entities[entity_type.__name__]
        source_columns = list([attribute.name for attribute in entity_sources_config.attributes])
        logger.debug(f'Source columns: {source_columns}')
        schema = entity_type.schema()
        schema_columns = list(schema['properties'].keys())
        logger.debug(f'Schema columns: {schema_columns}')
        invalid_columns = set(source_columns) - set(schema_columns)
        if invalid_columns:
            raise DataException(f'Unknown columns in source configuration: {invalid_columns}')
        return get_id_property(schema)

    def read_entity_data(self, entity_type) -> Sequence:
        """
        Reads data for an entity type from the source files that are specified
        in the sources config file.

        :param entity_type: the entity type, e.g., Individual.
        :return: A sequence of entities.
        """
        logger.info(f'* Reading {entity_type.__name__} data ...')
        if entity_type.__name__ not in self.sources_config.entities:
            raise DataException(f'No source configuration found for the {entity_type.__name__} entity')

        id_property = self.read_id_property(entity_type)

        entity_sources_config = self.sources_config.entities[entity_type.__name__]

        validate_derived_values_not_in_source_config(entity_type, entity_sources_config)

        source_files, source_file_id_mapping = get_source_files(entity_sources_config, id_property)

        # Read data from source files
        source_data = {}
        entity_data = {}
        for source_file in source_files:
            source_file_data = self.read_source_file_data(source_file)
            if len(source_file_data) == 0:
                raise DataException(f'No records in {source_file}')
            source_id_column = source_file_id_mapping[source_file]
            record_number = 0
            item_ids = set()
            for item in source_file_data:
                record_number += 1
                if source_id_column not in item.keys():
                    raise DataException(f'Identifier column \'{source_id_column}\' not found in file {source_file}. '
                                        f'Is the delimiter configured correctly in the sources config?')
                item_id = item.get(source_id_column, None)
                if item_id is None or item_id == '':
                    raise DataException(f'Empty identifier in {source_file} record number {record_number}')
                if item_id in item_ids:
                    raise DataException(f'Duplicate identifier in {source_file} record number {record_number}')
                item_ids.add(item_id)
                if item_id not in entity_data:
                    entity_data[item_id] = {id_property: item_id}
            source_data[source_file] = source_file_data

        logger.debug(f'{entity_type.__name__} entity data: {entity_data}')

        # Merge data from different sources files
        for attribute in entity_sources_config.attributes:
            if attribute.name == id_property:
                continue
            # add data to entities for attribute
            logger.debug(f'Adding data for attribute {attribute.name}')
            for source in attribute.sources:
                # default column name is the attribute name
                source_column = source.column if source.column is not None else attribute.name
                # check if column is in the source data
                first_record = source_data[source.file][0]
                if source_column not in first_record.keys():
                    raise DataException(f'Column \'{source_column}\' not found in file {source.file}. '
                                        f'Is the delimiter configured correctly in the sources config?')
                # add data from source to attribute
                logger.debug(
                    f'Adding data for attribute {attribute.name} from source {source.file}:{source_column}')
                source_id_column = source_file_id_mapping[source.file]
                for entity_id, entity in entity_data.items():
                    if entity.get(attribute.name) is not None:
                        continue
                    source_records = list([record for record in source_data[source.file]
                                           if record[source_id_column] == entity_id])
                    if not source_records:
                        continue
                    if len(source_records) > 1:
                        raise DataException(f'Multiple records for {entity_type.__name__}'
                                            f' with id {entity_id} in file {source.file}')
                    value = source_records[0][source_column]
                    if value == '':
                        value = None
                    if value is not None and source.date_format is not None:
                        try:
                            value = datetime.strptime(value, source.date_format)
                        except Exception as e:
                            logger.error(e)
                            raise DataException(f'Error parsing {attribute.name} from'
                                                f' source {source.file}:{source_column} with id {entity_id}')
                    entity[attribute.name] = value

        logger.debug(f'{entity_type.__name__} entity data: {entity_data}')

        try:
            return transform_entities(
                entity_data.values(),
                entity_type.schema(),
                lambda e: entity_type(**e)
            )
        except DataException as e:
            logger.error(f'Please check source files: {", ".join(source_files)}')
            raise e

    def read_subject_data(self) -> CentralSubjectRegistry:
        logger.info('Reading subject registry data ...')

        subject_registry_data: Dict[str, Sequence[SubjectEntity]] = {}

        for entity_type in list(SubjectEntity.__args__):
            subject_registry_data[entity_type.__name__] = self.read_entity_data(entity_type)

        return CentralSubjectRegistry.create(subject_registry_data)

    def read_study_data(self) -> StudyRegistry:
        logger.info('Reading study registry data ...')

        study_registry_data = {}

        for entity_type in list(StudyEntity.__args__):
            study_registry_data[entity_type.__name__] = self.read_entity_data(entity_type)

        return StudyRegistry.create(study_registry_data)



from typing import Dict, Sequence, List

from transmart_loader.transmart import DataCollection, Study, TrialVisit, Patient, DimensionType, ValueType, Modifier, \
    Dimension

from csr.csr import CentralSubjectRegistry, StudyRegistry, Individual, SubjectEntity
from csr2transmart.mappers.observation_mapper import ObservationMapper
from csr2transmart.mappers.ontology_mapper import OntologyMapper
from csr2transmart.ontology_config import TreeNode


class CsrMapper:
    """
    Map CSR data to Transmart Loader input format
    """
    def __init__(self, study_id: str, top_tree_node: str):
        self.default_trial_visit_label = 'GENERAL'
        self.patient_col = 'INDIVIDUAL_ID'
        self.study_id = study_id
        self.top_tree_node = top_tree_node
        self.individual_id_to_patient: Dict[str, Patient] = {}
        self.dimensions: List[Dimension] = []
        self.modifier_key_to_modifier: Dict[str, Modifier] = {}

    def map_study(self) -> Study:
        return Study(self.study_id, self.study_id)

    def map_default_trial_visit(self, study: Study) -> TrialVisit:
        return TrialVisit(study, self.default_trial_visit_label)

    def map_patients(self, individuals: Sequence[Individual]):
        for individual in individuals:
            patient = Patient(individual.individual_id, individual.gender, [])
            self.individual_id_to_patient[individual.individual_id] = patient

    def map_dimensions_and_modifiers(self):
        entities = list(SubjectEntity.__args__)
        entities.remove(Individual)
        for index, entity_type in enumerate(entities):
            type_name = entity_type.schema()['title']
            modifier = Modifier(type_name,
                                type_name,
                                type_name,
                                ValueType.Categorical)
            self.modifier_key_to_modifier[type_name] = modifier

            modifier_dimension = Dimension(type_name,
                                           modifier,
                                           DimensionType.Subject,
                                           index+2)
            self.dimensions.append(modifier_dimension)

    def map(self,
            subject_registry: CentralSubjectRegistry,
            study_registry: StudyRegistry,
            src_ontology: Sequence[TreeNode]) -> DataCollection:
        self.map_patients(subject_registry.individuals)
        study = self.map_study()
        default_trial_visit = self.map_default_trial_visit(study)
        self.map_dimensions_and_modifiers()

        ontology_mapper = OntologyMapper(self.top_tree_node)
        ontology = ontology_mapper.map(src_ontology)

        observation_mapper = ObservationMapper(default_trial_visit,
                                               self.individual_id_to_patient,
                                               ontology_mapper.concept_code_to_concept,
                                               self.modifier_key_to_modifier)
        observation_mapper.map_observations(subject_registry, study_registry)

        return DataCollection(ontology_mapper.concept_code_to_concept.values(),
                              self.modifier_key_to_modifier.values(),
                              self.dimensions,
                              [study],
                              [default_trial_visit],
                              [],
                              ontology,
                              self.individual_id_to_patient.values(),
                              observation_mapper.observations,
                              [],
                              [])



from typing import Dict, Any

from pydantic import BaseModel


class ColumnValueMapping(BaseModel):
    value_mapping: Dict[Any, Any]


class CodeBook(BaseModel):
    column_mappings: Dict[str, ColumnValueMapping]



import os
from typing import Optional, Sequence

from csr.tabular_file_reader import TabularFileReader
from sources2csr.ngs import NGS, LibraryStrategy
from sources2csr.ngs_reader import NgsReader, ReaderException


class NgsMafReader(NgsReader):
    """ Mutation data reader.
        Parses NGS files in the MAF format. The MAF files are gzipped.
    """

    def __init__(self, input_dir: str):
        super().__init__(input_dir, LibraryStrategy.SNV)
        self.sample_id_column_name = 'Tumor_Sample_Barcode'

    def read_data(self, filename: str) -> Optional[Sequence[NGS]]:
        """ Reads .maf.gz. file.
        Sample_id should be specified in the :attr:`self.sample_id_column_name` column.

        :param filename: name of the input file
        :return: Sequence of NGS objects
        """
        data = TabularFileReader(os.path.join(self.input_dir, filename)).read_data()
        biosource_biomaterial_dict = dict()
        if len(data) > 1:
            for row in data:
                try:
                    col_value = row[self.sample_id_column_name]
                except KeyError:
                    raise ReaderException("Invalid {} file. No column with name {}. Cannot read sample ids."
                                          .format(filename, self.sample_id_column_name))
                biosource_biomaterial = self.biosource_biomaterial_from_sample_id(col_value, filename)
                biosource_biomaterial_dict.setdefault(biosource_biomaterial[0], []).append(biosource_biomaterial[1])
        else:
            raise ReaderException("Cannot read NGS data from file: {}. Empty data.".format(filename))
        return self.map_ngs(biosource_biomaterial_dict, filename)



from datetime import date
from typing import Sequence, Optional, Union, Dict, List

from pydantic import BaseModel, Schema, validator

from csr.exceptions import DataException


class Individual(BaseModel):
    """
    Individual entity
    """
    individual_id: str = Schema(..., min_length=1, identity=True)
    taxonomy: Optional[str]
    gender: Optional[str]
    birth_date: Optional[date]
    death_date: Optional[date]
    ic_type: Optional[str]
    ic_version: Optional[float]
    ic_given_date: Optional[date]
    ic_withdrawn_date: Optional[date]
    ic_material: Optional[str]
    ic_data: Optional[str]
    ic_linking_ext: Optional[str]
    report_her_susc: Optional[str]
    report_inc_findings: Optional[str]
    diagnosis_count: Optional[int] = Schema(None, derived=True)
    age_first_diagnosis: Optional[int] = Schema(None, derived=True)


class Diagnosis(BaseModel):
    """
    Diagnosis entity
    """
    diagnosis_id: str = Schema(..., min_length=1, identity=True)
    individual_id: str = Schema(..., min_length=1, references='Individual')
    tumor_type: Optional[str]
    topography: Optional[str]
    treatment_protocol: Optional[str]
    tumor_stage: Optional[str]
    diagnosis_date: Optional[date]
    center_treatment: Optional[str]


class Biosource(BaseModel):
    """
    Biosource entity
    """
    biosource_id: str = Schema(..., min_length=1, identity=True)
    biosource_dedicated: Optional[str]
    individual_id: str = Schema(..., min_length=1, references='Individual')
    diagnosis_id: Optional[str] = Schema(None, min_length=1, references='Diagnosis')
    src_biosource_id: Optional[str] = Schema(None, min_length=1, references='Biosource')
    tissue: Optional[str]
    biosource_date: Optional[date]
    disease_status: Optional[str]
    tumor_percentage: Optional[int]

    @validator('src_biosource_id')
    def check_self_reference(cls, src_biosource_id, values):
        if src_biosource_id == values['biosource_id']:
            raise DataException(f'Biosource cannot be derived from itself')
        return src_biosource_id


class Biomaterial(BaseModel):
    """
    Biomaterial entity
    """
    biomaterial_id: str = Schema(..., min_length=1, identity=True)
    src_biosource_id: str = Schema(..., min_length=1, references='Biosource')
    src_biomaterial_id: Optional[str] = Schema(None, min_length=1, references='Biomaterial')
    biomaterial_date: Optional[date]
    type: Optional[str]
    library_strategy: Optional[List[str]] = Schema(None, derived=True)
    analysis_type: Optional[List[str]] = Schema(None, derived=True)

    @validator('src_biomaterial_id')
    def check_self_reference(cls, src_biomaterial_id, values):
        if src_biomaterial_id == values['biomaterial_id']:
            raise DataException(f'Biomaterial cannot be derived from itself')
        return src_biomaterial_id


class Study(BaseModel):
    """
    Study
    """
    study_id: str = Schema(..., min_length=1, identity=True)
    acronym: Optional[str]
    title: Optional[str]
    datadictionary: Optional[str]


class IndividualStudy(BaseModel):
    """
    Study to individual mapping
    """
    individual_study_id: int = Schema(..., identity=True)
    individual_id: str = Schema(..., min_length=1, references='Individual')
    study_id: str = Schema(..., min_length=1, references='Study')


SubjectEntity = Union[Individual, Diagnosis, Biosource, Biomaterial]


class CentralSubjectRegistry(BaseModel):
    """
    Central subject registry
    """
    individuals: Sequence[Individual]
    diagnoses: Optional[Sequence[Diagnosis]]
    biosources: Optional[Sequence[Biosource]]
    biomaterials: Optional[Sequence[Biomaterial]]

    @staticmethod
    def create(entity_data: Dict[str, Sequence[SubjectEntity]]):
        return CentralSubjectRegistry(
            individuals=entity_data['Individual'],
            diagnoses=entity_data['Diagnosis'],
            biosources=entity_data['Biosource'],
            biomaterials=entity_data['Biomaterial']
        )


StudyEntity = Union[Study, IndividualStudy]


class StudyRegistry(BaseModel):
    """
    Study registry
    """
    studies: Optional[Sequence[Study]]
    individual_studies: Optional[Sequence[IndividualStudy]]

    @staticmethod
    def create(entity_data: Dict[str, Sequence[StudyEntity]]):
        return StudyRegistry(
            studies=entity_data['Study'],
            individual_studies=entity_data['IndividualStudy']
        )



from csr.csr import StudyRegistry, Study, IndividualStudy
from csr.entity_writer import EntityWriter


class StudyRegistryWriter(EntityWriter):
    """Writer that writes study registry data to tab delimited files.
    """

    def __init__(self, output_dir: str):
        EntityWriter.__init__(self, output_dir)

    def write(self, study_registry: StudyRegistry):
        # Write studies
        self.write_entities('study.tsv', Study.schema(), study_registry.studies)
        # Write individual-study links
        self.write_entities('individual_study.tsv', IndividualStudy.schema(), study_registry.individual_studies)
